ERROR_TEMPLATE = "⚠ [bold red]{}[/bold red]"
COMMAND_UNDONE = "✋ [bold Yellow]Command will not take effect.[/bold Yellow]"
COMMAND_HINT = "👉 [bold Green]You may need to run:[/bold Green] [italic]{}[/italic]"
COMMAND_SUCCESSFUL = "[green]✔️ {}[green]"
OATH = "\n" \
       "    Night gathers, and now my watch begins. \n" \
       "    It shall not end until my death. \n" \
       "    I shall take no wife, hold no lands, father no children. \n" \
       "    I shall wear no crowns and win no glory. \n" \
       "    I shall live and die at my post.\n" \
       "    I am the sword in the darkness. \n" \
       "    I am the watcher on the walls. \n" \
       "    I am the fire that burns against the cold,\n\r" \
       "        the light that brings the dawn,\n\r" \
       "        the horn that wakes the sleepers,\n\r" \
       "        the shield that guards the realms of men. \n\r" \
       "    I pledge my life and honor to the Night’s Watch,\n\r" \
       "        for this night and all the nights to come. \n\r" \
       "    G.R.R. Martin\n\r"



"""This module provides the Automatic Deployer CLI."""

import json
import os
import time

import typer
from deployer import __app_name__, __version__
from deployer.consts import (
    ERROR_TEMPLATE,
    COMMAND_UNDONE,
    COMMAND_HINT,
    COMMAND_SUCCESSFUL,
    OATH
)
from deployer.manipulators import GitManipulator, CommandManipulator
from rich import print


def main(
        silent: bool = typer.Option(
            default=False,
            help="Execute the command without any output"
        )
):
    if not silent:
        print("Execution details...")
        state["verbose"] = True


def get_app_description() -> str:
    return f"""
    Automatic Deploy CLI app.

    App Name: {__app_name__}\n
    App version: {__version__}\n
    Programmer: Reza Mohammadi
    """


app = typer.Typer(callback=main, help=get_app_description())
project_root = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
state = {"verbose": True}


def get_configs() -> dict:
    # load configs
    app_root = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    with open(os.path.join(app_root, "data", "configs.json"), mode='r') as fh:
        app_configs = json.load(fh)

    return app_configs


def save_configs(new_configs: dict) -> None:
    # load configs
    app_root = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    with open(os.path.join(app_root, "data", "configs.json"), mode='w') as fh:
        json.dump(new_configs, fh, indent=4, ensure_ascii=False)


def get_oath() -> str:
    return OATH


@app.command()
def setup(
        git_url: str = typer.Option(
            default="",
            prompt="Enter git destination url to watch",
            help="Git Repository URL, "
                 "ex [http://gitlab.it/datamining/natural-language-understanding/opinion-news-agancy]"
        ),
        track_commits: bool = typer.Option(
            default=True,
            prompt="Enable tracking commits in watcher?",
        ),
        track_tags: bool = typer.Option(
            default=False,
            prompt="Enable tracking tags in watcher?",
        ),
        command: str = typer.Option(
            default="",
            prompt="Enter your command",
            help="Bash command, "
                 "ex [gunicorn app:app --workers 4 --worker-class uvicorn.workers.UvicornWorker --bind 0.0.0.0:8200]"
        ),
        run_after_setup: bool = typer.Option(
            default=False,
            prompt="Run One time after setup is completed?",
            help="This will not start the watch!"
        ),
        sleep_interval: int = typer.Option(
            default=2,
            prompt="Sleep interval before two checking",
            help="This parameter value should be considered as in minutes interval. "
                 "Set this parameter for preventing spamming the git server. "
                 "Set the parameter to 0 if it is crucial to check in real time"
        ),
):
    if git_url.strip() == "":
        print(ERROR_TEMPLATE.format("Git Repository URL can not be empty string"))
        print(COMMAND_UNDONE)
        raise typer.Exit()

    if command.strip() == "":
        print(ERROR_TEMPLATE.format("Command can not be empty string"))
        print(COMMAND_UNDONE)
        raise typer.Exit()

    # if both track_commits and track_tags are False
    if (not track_commits) and (not track_tags):
        print(ERROR_TEMPLATE.format("At least one of `track-commits` or `track-tags` should be set for watch"))
        print(COMMAND_UNDONE)
        raise typer.Exit()

    gtm = GitManipulator(git_url)
    is_valid = gtm.is_repo_valid()
    if not is_valid:
        print(ERROR_TEMPLATE.format("Invalid repo url!"))
        print(COMMAND_UNDONE)
        raise typer.Exit()

    if run_after_setup:
        gtm.setup_repo(project_root)
        cmm = CommandManipulator(run_as_root=True)
        cmm.run(message="Running Command", command=command, repo_directory=gtm.repo_directory, show_output=True)

    # report
    if state["verbose"]:
        print(f"[magenta]Git Repository URL:[/magenta] [blue]{git_url}[/blue]")
        print(f"[magenta]Execution Command:[/magenta] [blue]{command}[/blue]")
        print(f"[magenta]Sleep Interval:[/magenta] [blue]{sleep_interval}[/blue]")

    # save to file
    new_configs = {
        "setup_done": True,
        "git_repo_url": git_url,
        "execute_command": command,
        "sleep_interval": sleep_interval,
        "track_commits": track_commits,
        "track_tags": track_tags
    }
    save_configs(new_configs)
    print(COMMAND_SUCCESSFUL.format("Configurations updated!"))


@app.command()
def start(
        see_the_oath_and_metaphor_of_the_app: bool = True,
        verbose: bool = typer.Option(default=False, ),
):
    # one time fetching
    configs = get_configs()

    # check for setup done
    if not configs["setup_done"]:
        print(ERROR_TEMPLATE.format("Setup did not run successfully."))
        print(COMMAND_HINT.format("deployer setup"))
        raise typer.Exit()

    if see_the_oath_and_metaphor_of_the_app:
        print(get_oath())

    # clone repo
    gtm = GitManipulator(git_url=configs["git_repo_url"])

    print("Cleaning src folder...")

    print("Setting up repo handler...")
    gtm.setup_repo(project_root)

    print("Getting Commits and Tags...")
    latest_commits = gtm.fetch_commits()

    if configs["track_tags"]:
        latest_tags = gtm.fetch_tags(renew=True)
    else:
        latest_tags = gtm.fetch_tags(renew=False)

    # run the command
    cmm = CommandManipulator(run_as_root=True)
    cmm.run(
        message="running command",
        command=configs["execute_command"],
        repo_directory=gtm.repo_directory,
        show_output=verbose
    )

    sleep_duration = configs["sleep_interval"] * 60
    # watch the repo
    while True:
        print(f"Sleeping for {sleep_duration} seconds")
        time.sleep(sleep_duration)

        print("Fetching new changes...")
        commits = []
        if configs["track_commits"]:
            commits = gtm.fetch_commits(renew=False)

        tags = []
        if configs["track_tags"]:
            tags = gtm.fetch_tags(renew=True)

        if (commits != latest_commits) or (tags != latest_tags):
            typer.echo("Changes detected in source repository")
            latest_commits = commits
            latest_tags = tags

            typer.echo("Running command")
            cmm.run(
                message="running command",
                command=configs["execute_command"],
                repo_directory=gtm.repo_directory,
                show_output=verbose
            )


@app.command()
def reset(
        sure: bool = typer.Option(
            default=False,
            prompt="Are you sure?",
        ),
):
    if sure:
        new_configs = {
            "setup_done": False,
        }
        save_configs(new_configs)
        print(COMMAND_SUCCESSFUL.format("Configurations Reset!"))
    else:
        print(COMMAND_UNDONE)


@app.command()
def configs():
    print("deployer configurations")
    print(get_configs(), sep='\n')


@app.command()
def version():
    print(f"App Name: {__app_name__}\nApp version: {__version__}")



import os
import platform
import subprocess
import time


class CommandManipulator:
    wheel = ['|', '/', '-', '\\']

    def __init__(self, run_as_root: bool = False):
        if platform.system() == 'Linux' and run_as_root and os.geteuid() != 0:
            print('This script must be run as root')
            raise SystemExit(1)

    def run(self, message: str, command: str, repo_directory: str, die: bool = True, show_output: bool = False) -> bool:
        temp = '\r [%s] ' + message
        idx = 0

        if platform.system() == 'Windows':
            command = f"cd \"{repo_directory}\\src\"; " + command
            proc = subprocess.Popen(
                ['powershell.exe', command],
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )

        elif platform.system() == 'Linux':
            command = f"cd \"{repo_directory}/src\" && " + command
            proc = subprocess.Popen(
                command,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                executable='/bin/bash',
                text=True
            )

        while proc.poll() is None:
            if idx < len(self.wheel) - 1:
                idx += 1
            else:
                idx = 0

            print(temp % self.wheel[idx], end='')
            time.sleep(0.1)

        if proc.returncode == 0:
            print(temp % '+')

            if show_output:
                outs, errs = proc.communicate(timeout=15)
                outs_str = outs.decode('utf-8')
                print("Output:")
                print(outs_str.split("Mode")[0])

            return True

        print(temp % 'x', end=': ')
        try:
            print(proc.stderr.read().decode('utf-8'), end='')
        except:
            print(proc.stderr.read(), end='')

        if die:
            raise SystemExit(proc.returncode)

        return False



from tempfile import TemporaryDirectory

import git
import shutil
import os


class GitManipulator:
    repo_directory: str | None = None
    _git_url: str | None = None
    _repo: git.Repo | None = None
    _remote: git.Remote | None = None
    _branch: str | None = None

    def __init__(
            self,
            git_url: str,
    ):
        self._git_url = git_url

    def is_repo_valid(self) -> bool:
        try:
            with TemporaryDirectory() as dir_name:
                self._clone_repo(dir_name)
            return True
        except git.GitCommandError as ex:
            if (
                    "does not exist" in ex.stderr or
                    "not found" in ex.stderr or
                    "Logon failed" in ex.stderr
            ):
                return False
            else:
                raise ex

    def setup_repo(self, root_directory: str) -> None:
        self.repo_directory = root_directory
        self._clone_repo()
        self._branch = self._repo.heads[0].name
        self._remote = git.Remote(self._repo, "origin")

    def update_repo(self, renew: bool = False):
        if self.repo_directory is None:
            raise Exception("setup_repo first!")

        if renew:
            self._clone_repo()
        else:
            self._remote.pull(refspec=self._branch)

    def fetch_commits(
            self,
            renew: bool = False,
    ) -> list[dict]:
        self.update_repo(renew)
        commits_list = []
        for c in self._repo.iter_commits():
            commits_list.append(
                {
                    "committed_date": c.committed_date,
                    "committed_datetime": c.committed_datetime,
                    "summary": c.summary,
                }
            )
        return commits_list

    def fetch_tags(
            self,
            renew: bool = False
    ) -> list[dict]:
        self.update_repo(renew)
        tags_list = []
        for t in self._repo.tags:
            tags_list.append(
                {
                    "name": t.name
                }
            )
        return tags_list

    def _clone_repo(self, root_directory: str | None = None) -> None:
        if root_directory is None:
            root_directory = self._get_cleaned_src_folder()

        self._repo = git.Repo.clone_from(url=self._git_url, to_path=root_directory)

    def _get_cleaned_src_folder(self) -> str:
        def onerror(func, path, exc_info):
            """
            Error handler for ``shutil.rmtree``.

            If the error is due to an access error (read only file)
            it attempts to add write permission and then retries.

            If the error is for another reason it re-raises the error.

            Usage : ``shutil.rmtree(path, onerror=onerror)``
            """
            import stat
            # Is the error an access error?
            if not os.access(path, os.W_OK):
                os.chmod(path, stat.S_IWUSR)
                func(path)
            else:
                raise

        src_path = os.path.join(self.repo_directory, "src")
        # os.remove(src_path)
        try:
            shutil.rmtree(src_path, onerror=onerror)
            os.mkdir(src_path)
        except:
            pass

        return src_path
